"""
This package contains code that was "generated" via metaprogramming.

This happens, because Python is not flexible enough to do most tasks
common in typed functional programming.

Policy:

1. We store implementations in regular ``.py`` files.
2. We store generated type annotations in ``.pyi`` files.
3. We re-export these functions into regular modules as public values.

Please, do not touch this code unless you know what you are doing.
"""
